@@ -1,6 +1,6 @@
 /* ncftpget.c
  *
- * Copyright (c) 1996-2001 Mike Gleason, NCEMRSoft.
+ * Copyright (c) 1996-2005 Mike Gleason, NcFTP Software.
  * All rights reserved.
  *
  * A non-interactive utility to grab files from a remote FTP server.
@@ -8,25 +8,20 @@
  */
 
 #include "syshdrs.h"
+#ifdef PRAGMA_HDRSTOP
+#	pragma hdrstop
+#endif
 
-#if defined(WIN32) || defined(_WINDOWS)
-#	include "..\ncftp\getopt.h"
-#	define getopt Getopt
-#	define optarg gOptArg
-#	define optind gOptInd
-	WSADATA wsaData;
-	int wsaInit = 0;
-
-	__inline void DisposeWinsock(int aUNUSED) { if (wsaInit > 0) WSACleanup(); wsaInit--; }
+#if (defined(WIN32) || defined(_WINDOWS)) && !defined(__CYGWIN__)
 #	include "..\ncftp\util.h"
 #	include "..\ncftp\spool.h"
 #	include "..\ncftp\pref.h"
-#	include "..\ncftp\getline.h"
+#	include "..\ncftp\gl_getline.h"
 #else
 #	include "../ncftp/util.h"
 #	include "../ncftp/spool.h"
 #	include "../ncftp/pref.h"
-#	include "../ncftp/getline.h"
+#	include "../ncftp/gl_getline.h"
 #endif
 
 #include "gpshare.h"
@@ -42,10 +37,7 @@ extern char gFirewallPass[32];
 extern unsigned int gFirewallPort;
 extern char gFirewallExceptionList[256];
 extern int gFwDataPortMode;
-extern char gOS[], gVersion[];
-
-extern char *optarg;
-extern int optind;
+extern const char gOS[], gVersion[];
 
 static void
 Usage(void)
@@ -55,11 +47,12 @@ Usage(void)
 	fp = OpenPager();
 	(void) fprintf(fp, "NcFTPGet %.5s\n\n", gVersion + 11);
 	(void) fprintf(fp, "Usages:\n");
-	(void) fprintf(fp, "  ncftpget [flags] remote-host local-dir remote-path-names...   (mode 1)\n");
-	(void) fprintf(fp, "  ncftpget -f login.cfg [flags] local-dir remote-path-names...  (mode 2)\n");
-	(void) fprintf(fp, "  ncftpget [flags] ftp://url.style.host/path/name               (mode 3)\n");
-	(void) fprintf(fp, "  ncftpget -c [flags] remote-host remote-path-name > stdout (mode 4)\n");
-	(void) fprintf(fp, "  ncftpget -c [flags] ftp://url.style.host/path/name > stdout (mode 5)\n");
+	(void) fprintf(fp, "  ncftpget [flags] remote-host local-dir remote-path-names...      (mode 1)\n");
+	(void) fprintf(fp, "  ncftpget -f login.cfg [flags] local-dir remote-path-names...     (mode 2)\n");
+	(void) fprintf(fp, "  ncftpget [flags] ftp://url.style.host/path/name                  (mode 3)\n");
+	(void) fprintf(fp, "  ncftpget -c [flags] remote-host remote-path-name > stdout        (mode 4)\n");
+	(void) fprintf(fp, "  ncftpget -C [flags] remote-host remote-path-name local-path-name (mode 5)\n");
+	(void) fprintf(fp, "  ncftpget -c [flags] ftp://url.style.host/path/name > stdout      (mode 6)\n");
 	(void) fprintf(fp, "\nFlags:\n\
   -u XX  Use username XX instead of anonymous.\n\
   -p XX  Use password XX with the username.\n\
@@ -70,9 +63,12 @@ Usage(void)
   -t XX  Timeout after XX seconds.\n\
   -v/-V  Do (do not) use progress meters.\n\
   -f XX  Read the file XX for host, user, and password information.\n\
+  -h XX  Connect to host XX.  Useful for overriding host in -f config.file.\n\
+  -c     Read from remote host and write locally to stdout.\n\
+  -C     Read from remote host and write locally to specified file.\n\
   -A     Append to local files, instead of overwriting them.\n");
 	(void) fprintf(fp, "\
-  -z/-Z  Do (do not) not try to resume downloads (default: -z).\n\
+  -z/-Z  Do (do not) try to resume downloads (default: -z).\n\
   -E     Use regular (PORT) data connections.\n\
   -F     Use passive (PASV) data connections (default).\n\
   -DD    Delete remote file after successfully downloading it.\n\
@@ -81,6 +77,7 @@ Usage(void)
 	(void) fprintf(fp, "\
   -B XX  Try setting the SO_RCVBUF size to XX.\n\
   -r XX  Redial XX times until connected.\n\
+  -o XX  Specify miscellaneous options (see documentation).\n\
   -W XX  Send raw FTP command XX after logging in.\n\
   -X XX  Send raw FTP command XX after each file transferred.\n\
   -Y XX  Send raw FTP command XX before logging out.\n\
@@ -99,11 +96,11 @@ Usage(void)
   ncftpget -a -d /tmp/debug.log -t 60 ftp.wustl.edu . '/pub/README*'\n");
 
 	(void) fprintf(fp, "\nLibrary version: %s.\n", gLibNcFTPVersion + 5);
-	(void) fprintf(fp, "\nThis is a freeware program by Mike Gleason (mgleason@ncftp.com).\n");
-	(void) fprintf(fp, "This was built using LibNcFTP (http://www.ncftp.com/libncftp).\n");
+	(void) fprintf(fp, "\nThis is a freeware program by Mike Gleason (http://www.ncftp.com).\n");
+	(void) fprintf(fp, "This was built using LibNcFTP (http://www.ncftp.com/libncftp/).\n");
 
 	ClosePager(fp);
-	DisposeWinsock(0);
+	DisposeWinsock();
 	exit(kExitUsage);
 }	/* Usage */
 
@@ -111,10 +108,9 @@ Usage(void)
 
 
 static void
-Abort(int UNUSED(sigNum))
+Abort(int sigNum)
 {
-	LIBNCFTP_USE_VAR(sigNum);
-	signal(SIGINT, Abort);
+	signal(sigNum, Abort);
 
 	/* Hopefully the I/O operation in progress
 	 * will complete, and we'll abort before
@@ -127,14 +123,14 @@ Abort(int UNUSED(sigNum))
 	 * next ^C abends the program.
 	 */
 	if (fi.cancelXfer >= 2)
-		signal(SIGINT, SIG_DFL);
+		signal(sigNum, SIG_DFL);
 }	/* Abort */
 
 
 
 
 static int 
-Copy(FTPCIPtr cip, char *dstdir, const char ** volatile files, int rflag, int xtype, int resumeflag, int appendflag, int deleteflag, int tarflag, const char *const perfilecmd)
+Copy(FTPCIPtr cip, char *dstdir, const char **files, int rflag, int xtype, int resumeflag, int appendflag, int deleteflag, int tarflag, const char *const perfilecmd)
 {
 	int i;
 	int result;
@@ -145,7 +141,7 @@ Copy(FTPCIPtr cip, char *dstdir, const char ** volatile files, int rflag, int xt
 		file = files[i];
 		if (file == NULL)
 			break;
-		result = FTPGetFiles3(cip, file, dstdir, rflag, kGlobYes, xtype, resumeflag, appendflag, deleteflag, tarflag, NoConfirmResumeDownloadProc, 0);
+		result = FTPGetFiles3(cip, file, dstdir, rflag, kGlobYes, xtype, resumeflag, appendflag, deleteflag, tarflag, kNoFTPConfirmResumeDownloadProc, 0);
 		if (result != 0) {
 			FTPPerror(cip, result, kErrCouldNotStartDataTransfer, "ncftpget", file);
 			if (result != kErrLocalSameAsRemote) {
@@ -162,27 +158,27 @@ Copy(FTPCIPtr cip, char *dstdir, const char ** volatile files, int rflag, int xt
 
 
 
-int
+main_void_return_t
 main(int argc, char **argv)
 {
 	int result, c;
-	volatile int rflag = 0;
-	volatile int xtype = kTypeBinary;
-	volatile int appendflag = kAppendNo;
-	volatile int resumeflag = kResumeYes;
-	volatile int deleteflag = kDeleteNo;
-	volatile int tarflag = kTarYes;
+	int rflag = 0;
+	int xtype = kTypeBinary;
+	int appendflag = kAppendNo;
+	int resumeflag = kResumeYes;
+	int deleteflag = kDeleteNo;
+	int tarflag = kTarYes;
 	int progmeters;
 	char *dstdir = NULL;
+	char *dstlfile = NULL;
 	const char **flist;
-	const char *errstr;
-	volatile ExitStatus es;
+	ExitStatus es;
 	char url[512];
 	char urlfile[256];
 	char urldir[512];
 	int urlxtype;
-	LineList cdlist;
-	LinePtr lp;
+	FTPLineList cdlist;
+	FTPLinePtr lp;
 	int rc;
 	int nD = 0;
 	int batchmode = 0;
@@ -192,6 +188,7 @@ main(int argc, char **argv)
 	char *urlfilep;
 	const char *urldirp;
 	char precmd[320], postcmd[320], perfilecmd[320];
+	GetoptInfo opt;
 
 	InitWinsock();
 #if (defined(SOCKS)) && (SOCKS >= 5)
@@ -212,7 +209,7 @@ main(int argc, char **argv)
 	}
 
 	InitUserInfo();
-	fi.dataPortMode = kPassiveMode;
+	fi.dataPortMode = kFallBackToSendPortMode;
 	LoadFirewallPrefs(0);
 	if (gFwDataPortMode >= 0)
 		fi.dataPortMode = gFwDataPortMode;
@@ -230,54 +227,85 @@ main(int argc, char **argv)
 	postcmd[0] = '\0';
 	perfilecmd[0] = '\0';
 
-	while ((c = getopt(argc, argv, "P:u:j:p:e:d:t:aRTr:vVf:ADzZEFbcB:W:X:Y:")) > 0) switch(c) {
+	GetoptReset(&opt);
+	while ((c = Getopt(&opt, argc, argv, "P:u:j:p:h:e:d:t:aRTr:vVf:ADzZEFbcCB:W:X:Y:")) > 0) {
+		if (c == 'b') {
+			batchmode++;
+		}
+	}
+
+	if (batchmode > 0) {
+		GetoptReset(&opt);
+		while ((c = Getopt(&opt, argc, argv, "P:u:j:p:h:e:d:U:t:mar:RvVf:o:AT:S:EFcCyZzDbB:W:X:Y:")) > 0) switch(c) {
+			case 'v': case 'V': case 'A': case 'B': case 'T':
+			case 'd': case 'e': case 't': case 'r': case 'c':
+			case 'z': case 'Z': case 'C':
+				(void) fprintf(stderr, "The \"-%c\" option is not valid when used with conjunction with \"-%c\".\n", c, 'b');
+				exit(kExitUsage);
+				break;
+		}
+	}
+
+	GetoptReset(&opt);
+	while ((c = Getopt(&opt, argc, argv, "P:u:j:p:h:e:d:t:aRTr:vVf:o:ADzZEFbcCB:W:X:Y:")) > 0) switch(c) {
 		case 'P':
-			fi.port = atoi(optarg);	
+			fi.port = atoi(opt.arg);	
 			break;
 		case 'u':
-			(void) STRNCPY(fi.user, optarg);
-			memset(optarg, '*', strlen(fi.user));
+			(void) STRNCPY(fi.user, opt.arg);
+			memset(opt.arg, 0, strlen(fi.user));
+			opt.arg[0] = '?';
 			break;
 		case 'j':
-			(void) STRNCPY(fi.acct, optarg);
-			memset(optarg, '*', strlen(fi.acct));
+			(void) STRNCPY(fi.acct, opt.arg);
+			memset(opt.arg, 0, strlen(fi.acct));
+			opt.arg[0] = '?';
 			break;
 		case 'p':
-			(void) STRNCPY(fi.pass, optarg);	/* Don't recommend doing this! */
-			memset(optarg, '*', strlen(fi.pass));
+			(void) STRNCPY(fi.pass, opt.arg);	/* Don't recommend doing this! */
+			fi.leavePass = 1;
+			if (fi.pass[0] == '\0')
+				fi.passIsEmpty = 1;
+			memset(opt.arg, 0, strlen(fi.pass));
+			opt.arg[0] = '?';
+			break;
+		case 'h':
+			(void) STRNCPY(fi.host, opt.arg);
+			memset(opt.arg, 0, strlen(fi.user));
+			opt.arg[0] = '?';
 			break;
 		case 'e':
-			if (strcmp(optarg, "stdout") == 0)
+			if (strcmp(opt.arg, "stdout") == 0)
 				fi.errLog = stdout;
-			else if (optarg[0] == '-')
+			else if (opt.arg[0] == '-')
 				fi.errLog = stdout;
-			else if (strcmp(optarg, "stderr") == 0)
+			else if (strcmp(opt.arg, "stderr") == 0)
 				fi.errLog = stderr;
 			else
-				fi.errLog = fopen(optarg, FOPEN_APPEND_TEXT);
+				fi.errLog = fopen(opt.arg, FOPEN_APPEND_TEXT);
 			break;
 		case 'D':
 			/* Require two -D's in case they typo. */
 			nD++;
 			break;
 		case 'd':
-			if (strcmp(optarg, "stdout") == 0)
+			if (strcmp(opt.arg, "stdout") == 0)
 				fi.debugLog = stdout;
-			else if (optarg[0] == '-')
+			else if (opt.arg[0] == '-')
 				fi.debugLog = stdout;
-			else if (strcmp(optarg, "stderr") == 0)
+			else if (strcmp(opt.arg, "stderr") == 0)
 				fi.debugLog = stderr;
 			else
-				fi.debugLog = fopen(optarg, FOPEN_APPEND_TEXT);
+				fi.debugLog = fopen(opt.arg, FOPEN_APPEND_TEXT);
 			break;
 		case 't':
-			SetTimeouts(&fi, optarg);
+			SetTimeouts(&fi, opt.arg);
 			break;
 		case 'a':
 			xtype = kTypeAscii;
 			break;
 		case 'r':
-			SetRedial(&fi, optarg);
+			SetRedial(&fi, opt.arg);
 			break;
 		case 'R':
 			rflag = 1;
@@ -292,7 +320,10 @@ main(int argc, char **argv)
 			progmeters = 0;
 			break;
 		case 'f':
-			ReadConfigFile(optarg, &fi);
+			ReadConfigFile(opt.arg, &fi);
+			break;
+		case 'o':
+			fi.manualOverrideFeatures = opt.arg;
 			break;
 		case 'A':
 			appendflag = kAppendYes;
@@ -310,40 +341,48 @@ main(int argc, char **argv)
 			fi.dataPortMode = kPassiveMode;
 			break;
 		case 'b':
-			batchmode++;
+			/* handled above */
 			break;
 		case 'B':
-			fi.dataSocketRBufSize = (size_t) atol(optarg);	
+			fi.dataSocketRBufSize = (size_t) atol(opt.arg);	
 			break;
 		case 'W':
-			STRNCAT(precmd, optarg);
+			STRNCAT(precmd, opt.arg);
 			STRNCAT(precmd, "\n");
 			break;
 		case 'X':
-			STRNCAT(perfilecmd, optarg);
+			STRNCAT(perfilecmd, opt.arg);
 			STRNCAT(perfilecmd, "\n");
 			break;
 		case 'Y':
-			STRNCAT(postcmd, optarg);
+			STRNCAT(postcmd, opt.arg);
 			STRNCAT(postcmd, "\n");
 			break;
 		case 'c':
 			ftpcat = 1;
 			break;
+		case 'C':
+			ftpcat = 2;
+			break;
 		default:
 			Usage();
 	}
-	if (optind > argc - 1)
-		Usage();
-
-	if ((ftpcat != 0) && (batchmode != 0))
+	if (opt.ind > argc - 1)
 		Usage();
 
 	if (progmeters != 0)
 		fi.progress = PrStatBar;
 
+	if ((rflag != 0) && (ftpcat != 0)) {
+		fprintf(stderr, "The -R flag is not supported with -c/-C.\n\n");
+		Usage();
+	}
+
+	if ((perfilecmd[0] != '\0') && (rflag != 0))
+		(void) fprintf(stderr, "Warning: your -X command is only applied once per command-line parameter, and not for each file in the directory.\n");
+
 	if (fi.host[0] == '\0') {
-		(void) STRNCPY(url, argv[optind]);
+		(void) STRNCPY(url, argv[opt.ind]);
 		rc = FTPDecodeURL(&fi, url, &cdlist, urlfile, sizeof(urlfile), (int *) &urlxtype, NULL);
 		if (rc == kMalformedURL) {
 			(void) fprintf(stderr, "Malformed URL: %s\n", url);
@@ -351,22 +390,29 @@ main(int argc, char **argv)
 		} else if (rc == kNotURL) {
 			/* This is what should happen most of the time. */
 			if (ftpcat == 0) {
-				if (optind > argc - 3)
+				if (opt.ind > argc - 3)
 					Usage();
-				(void) STRNCPY(fi.host, argv[optind]);
-				dstdir = StrDup(argv[optind + 1]);
+				(void) STRNCPY(fi.host, argv[opt.ind]);
+				dstdir = StrDup(argv[opt.ind + 1]);
 				if (dstdir == NULL) {
 					(void) fprintf(stderr, "Out of memory?\n");
 					exit(kExitNoMemory);
 				}
 				StrRemoveTrailingLocalPathDelim(dstdir);
-				flist = (const char **) argv + optind + 2;
-			} else {
-				if (optind > argc - 2)
+				flist = (const char **) argv + opt.ind + 2;
+			} else if (ftpcat == 2) {
+				if (opt.ind > argc - 3)
 					Usage();
-				(void) STRNCPY(fi.host, argv[optind]);
+				(void) STRNCPY(fi.host, argv[opt.ind]);
 				dstdir = NULL;
-				flist = (const char **) argv + optind + 1;
+				flist = (const char **) argv + opt.ind + 1;
+				dstlfile = argv[opt.ind + 2];
+			} else {
+				if (opt.ind > argc - 2)
+					Usage();
+				(void) STRNCPY(fi.host, argv[opt.ind]);
+				dstdir = NULL;
+				flist = (const char **) argv + opt.ind + 1;
 			}
 		} else {
 			/* URL okay */
@@ -378,6 +424,12 @@ main(int argc, char **argv)
 				exit((int) es);
 			}
 
+			if (ftpcat == 2) {
+				if (opt.ind > argc - 2)
+					Usage();
+				dstlfile = argv[opt.ind + 1];
+			}
+
 			/* Allow "-a" flag to use ASCII mode
 			 * with the URL, since most people
 			 * don't know there is way to specify
@@ -389,25 +441,36 @@ main(int argc, char **argv)
 	} else {
 		/* login.cfg being used, so no remote-host argument. */
 		if (ftpcat == 0) {
-			if (optind > argc - 2)
+			if (opt.ind > argc - 2)
 				Usage();
-			dstdir = StrDup(argv[optind + 0]);
+			dstdir = StrDup(argv[opt.ind + 0]);
 			if (dstdir == NULL) {
 				(void) fprintf(stderr, "Out of memory?\n");
 				exit(kExitNoMemory);
 			}
 			StrRemoveTrailingLocalPathDelim(dstdir);
-			flist = (const char **) argv + optind + 1;
+			flist = (const char **) argv + opt.ind + 1;
+		} else if (ftpcat == 2) {
+			if (opt.ind > argc - 3)
+				Usage();
+			dstdir = StrDup(argv[opt.ind + 0]);
+			if (dstdir == NULL) {
+				(void) fprintf(stderr, "Out of memory?\n");
+				exit(kExitNoMemory);
+			}
+			StrRemoveTrailingLocalPathDelim(dstdir);
+			flist = (const char **) argv + opt.ind + 1;
+			dstlfile = argv[opt.ind + 2];
 		} else {
-			if (optind > argc - 1)
+			if (opt.ind > argc - 1)
 				Usage();
 			dstdir = NULL;
-			flist = (const char **) argv + optind + 0;
+			flist = (const char **) argv + opt.ind + 0;
 		}
 	}
 
 	if (strcmp(fi.user, "anonymous") && strcmp(fi.user, "ftp")) {
-		if (fi.pass[0] == '\0') {
+		if ((fi.pass[0] == '\0') && (fi.passIsEmpty == 0)) {
 			(void) gl_getpass("Password: ", fi.pass, sizeof(fi.pass));
 		}
 	}
@@ -461,6 +524,8 @@ main(int argc, char **argv)
 				NULL,
 				NULL,
 				(time_t) 0,	/* when: now */
+				0,
+				fi.manualOverrideFeatures,
 				0
 			);
 			if ((rc == 0) && (batchmode < 3)) {
@@ -471,7 +536,7 @@ main(int argc, char **argv)
 			/* List of files specified */
 			for (i=0; flist[i] != NULL; i++) {
 				STRNCPY(urlfile, flist[i]);
-				urlfilep = strrchr(urlfile, '/');
+				urlfilep = StrRFindLocalPathDelim(urlfile);
 				if (urlfilep == NULL) {
 					urldirp = ".";
 					urlfilep = urlfile;
@@ -504,6 +569,8 @@ main(int argc, char **argv)
 					NULL,
 					NULL,
 					(time_t) 0,	/* when: now */
+					0,
+					fi.manualOverrideFeatures,
 					0
 				);
 				if ((rc == 0) && (batchmode < 3)) {
@@ -516,66 +583,78 @@ main(int argc, char **argv)
 			if (batchmode == 1) {
 				RunBatch();
 			}
-			DisposeWinsock(0);
+			DisposeWinsock();
 			exit(kExitSuccess);
 		}
-		DisposeWinsock(0);
+		DisposeWinsock();
 		exit(kExitSpoolFailed);
 	}
 	
 	es = kExitOpenTimedOut;
-	errstr = "could not open remote host";
 	if ((result = FTPOpenHost(&fi)) < 0) {
 		(void) fprintf(stderr, "ncftpget: cannot open %s: %s.\n", fi.host, FTPStrError(result));
 		es = kExitOpenFailed;
-		DisposeWinsock(0);
+		DisposeWinsock();
 		exit((int) es);
 	}
 	if (fi.hasCLNT != kCommandNotAvailable)
 		(void) FTPCmd(&fi, "CLNT NcFTPGet %.5s %s", gVersion + 11, gOS);
 
-	errstr = "could not run pre-command remote host";
 	(void) AdditionalCmd(&fi, precmd, NULL);
 
 	if (flist == NULL) {
 		/* URL mode */
-		errstr = "could not change directory on remote host";
 		es = kExitChdirTimedOut;
 		if ((rc = FTPChdirList(&fi, &cdlist, NULL, 0, (kChdirFullPath|kChdirOneSubdirAtATime))) != 0) {
 			FTPPerror(&fi, rc, kErrCWDFailed, "ncftpget: Could not change directory", NULL);
 			(void) FTPCloseHost(&fi);
 			es = kExitChdirFailed;
-			DisposeWinsock(0);
+			DisposeWinsock();
 			exit((int) es);
 		}
 		
-		errstr = "could not read file from remote host";
 		es = kExitXferTimedOut;
 		(void) signal(SIGINT, Abort);
-		if (ftpcat != 0) {
-			if (FTPGetOneFile3(&fi, urlfile, NULL, xtype, STDOUT_FILENO, resumeflag, kAppendNo, deleteflag, NoConfirmResumeDownloadProc, 0) == kNoErr) {
+		if (ftpcat == 2) {
+			if (FTPGetOneFile3(&fi, urlfile, dstlfile, xtype, (-1), resumeflag, appendflag, deleteflag, kNoFTPConfirmResumeDownloadProc, 0) == kNoErr) {
+				es = kExitSuccess;
+			} else {
+				FTPPerror(&fi, rc, kErrCouldNotStartDataTransfer, "ncftpget", NULL);
+				es = kExitXferFailed;
+			}
+		} else if (ftpcat != 0) {
+			if (FTPGetOneFile3(&fi, urlfile, NULL, xtype, STDOUT_FILENO, resumeflag, kAppendNo, deleteflag, kNoFTPConfirmResumeDownloadProc, 0) == kNoErr) {
 				es = kExitSuccess;
 			} else {
 				FTPPerror(&fi, rc, kErrCouldNotStartDataTransfer, "ncftpget", NULL);
 				es = kExitXferFailed;
 			}
 		} else {
-			if ((rc = FTPGetFiles3(&fi, urlfile, ".", rflag, kGlobYes, xtype, resumeflag, appendflag, deleteflag, tarflag, NoConfirmResumeDownloadProc, 0)) < 0) {
-				FTPPerror(&fi, rc, kErrCouldNotStartDataTransfer, "ncftpget", NULL);
-				es = kExitXferFailed;
+			if ((rc = FTPGetFiles3(&fi, urlfile, ".", rflag, kGlobYes, xtype, resumeflag, appendflag, deleteflag, tarflag, kNoFTPConfirmResumeDownloadProc, 0)) < 0) {
+				if (rc == kErrLocalSameAsRemote) {
+					/* Show the message, but do not err-out. */
+					FTPPerror(&fi, rc, kErrCouldNotStartDataTransfer, "ncftpget", NULL);
+					es = kExitSuccess;
+				} else {
+					FTPPerror(&fi, rc, kErrCouldNotStartDataTransfer, "ncftpget", NULL);
+					es = kExitXferFailed;
+				}
 			} else {
 				es = kExitSuccess;
 
-				errstr = "could not run per-file-command remote host";
 				(void) AdditionalCmd(&fi, perfilecmd, urlfile);
 			}
 		}
 	} else {
-		errstr = "could not read file from remote host";
 		es = kExitXferTimedOut;
 		(void) signal(SIGINT, Abort);
-		if (ftpcat != 0) {
-			if (FTPGetOneFile3(&fi, flist[0], NULL, xtype, STDOUT_FILENO, resumeflag, kAppendNo, deleteflag, NoConfirmResumeDownloadProc, 0) == kNoErr)
+		if (ftpcat == 2) {
+			if (FTPGetOneFile3(&fi, flist[0], dstlfile, xtype, (-1), resumeflag, appendflag, deleteflag, kNoFTPConfirmResumeDownloadProc, 0) == kNoErr)
+				es = kExitSuccess;
+			else
+				es = kExitXferFailed;
+		} else if (ftpcat != 0) {
+			if (FTPGetOneFile3(&fi, flist[0], NULL, xtype, STDOUT_FILENO, resumeflag, kAppendNo, deleteflag, kNoFTPConfirmResumeDownloadProc, 0) == kNoErr)
 				es = kExitSuccess;
 			else
 				es = kExitXferFailed;
@@ -587,11 +666,10 @@ main(int argc, char **argv)
 		}
 	}
 
-	errstr = "could not run post-command remote host";
 	(void) AdditionalCmd(&fi, postcmd, NULL);
 	
 	(void) FTPCloseHost(&fi);
-	DisposeWinsock(0);
+	DisposeWinsock();
 	
 	exit((int) es);
 }	/* main */
